# Copyright 2019 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
Samplers that handle binary polynomials: problems with binary variables that are
not constrained to quadratic interactions.

"""

import abc
import warnings

import typing

from dimod.core.composite import Composite
from dimod.higherorder.polynomial import BinaryPolynomial
from dimod.sampleset import SampleSet

from dimod.typing import Bias, Variable

__all__ = 'PolySampler', 'ComposedPolySampler'


class PolySampler(abc.ABC):
    """Sampler that supports binary polynomials.

    Binary polynomials are an extension of binary quadratic models that allow
    higher-order interactions.

    """
    @abc.abstractproperty  # for python2 compatibility
    def parameters(self):
        """dict: A dict where keys are the keyword parameters accepted by the sampler
        methods and values are lists of the properties relevant to each parameter.
        """
        pass

    @abc.abstractproperty  # for python2 compatibility
    def properties(self):
        """dict: A dict containing any additional information about the sampler.
        """
        pass

    @abc.abstractmethod
    def sample_poly(self, polynomial, **kwargs):
        """Sample from a higher-order polynomial."""
        pass

    def sample_hising(self, h: typing.Mapping[Variable, Bias],
                      J: typing.Mapping[typing.Tuple[Variable, Variable], Bias], **kwargs) -> SampleSet:
        """Sample from a higher-order Ising model.

        Converts the given higher-order Ising model to a :obj:`.BinaryPolynomial`
        and calls :meth:`.sample_poly`.

        Args:
            h: Variable biases of the Ising problem.
            J: Interaction biases of the Ising problem.

            **kwargs:
                See :meth:`.sample_poly` for additional keyword definitions.

        Returns:
            Samples from the higher-order Ising model.

        See also:
            :meth:`.sample_poly`, :meth:`.sample_hubo`

        """
        return self.sample_poly(BinaryPolynomial.from_hising(h, J), **kwargs)

    def sample_hubo(self, H: typing.Mapping[typing.Tuple[Variable, Variable], Bias],
                    **kwargs) -> SampleSet:
        """Sample from a higher-order unconstrained binary optimization problem.

        Converts the given higher-order unconstrained binary optimization
        problem to a :obj:`.BinaryPolynomial` and then calls :meth:`.sample_poly`.

        Args:
            H: Coefficients of the HUBO.

            **kwargs:
                See :meth:`.sample_poly` for additional keyword definitions.

        Returns:
            Samples from a higher-order unconstrained binary optimization problem.

        See also:
            :meth:`.sample_poly`, :meth:`.sample_hising`

        """
        return self.sample_poly(BinaryPolynomial.from_hubo(H), **kwargs)


class ComposedPolySampler(PolySampler, Composite):
    """Abstract base class for dimod composed polynomial samplers.

    Inherits from :class:`.PolySampler` and :class:`.Composite`."""
    pass



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
A sampler that gives random samples.
"""

from dimod.core.sampler import Sampler
from dimod.reference.samplers.identity_sampler import IdentitySampler

__all__ = ['RandomSampler']

class RandomSampler(Sampler):
    """A sampler that gives random samples for testing.

    Examples:
        This example produces 10 samples for a two-variable problem.

        >>> bqm = dimod.BinaryQuadraticModel.from_qubo({('a', 'b'): 1})
        >>> sampler = dimod.RandomSampler()
        >>> sampleset = sampler.sample(bqm, num_reads=10)
        >>> len(sampleset)
        10
    """
    properties = None

    parameters: dict = None
    """Keyword arguments accepted by the sampling methods.

    Contents are a dict with the following keys: ``{'num_reads': []}``.
    """

    def __init__(self):
        self.parameters = {'num_reads': []}
        self.properties = {}

    def sample(self, bqm, *, num_reads=10, seed=None, **kwargs):
        """Return random samples for a binary quadratic model.

        Variable assignments are chosen by coin flip.

        Args:
            bqm (:obj:`.BinaryQuadraticModel`):
                Binary quadratic model to be sampled from.

            num_reads (int, optional, default=10):
                Number of reads.

            seed (int (32-bit unsigned integer), optional):
                Seed to use for the PRNG. Specifying a particular seed with a
                constant set of parameters produces identical results. If not
                provided, a random seed is chosen.

        Returns:
            :obj:`.SampleSet`

        """
        # as an implementation detail, we can use IdentitySampler here, but
        # in order to save on future changes that decouple them, we won't
        # subclass
        return IdentitySampler().sample(bqm, num_reads=num_reads, seed=seed,
                                        initial_states_generator='random',
                                        **kwargs)



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import collections.abc as abc

import dimod


def assert_sampler_api(sampler):
    """Assert that an instantiated sampler exposes correct properties and methods.

    Args:
        sampler (:obj:`.Sampler`):
            User-made dimod sampler.

    Raises:
        AssertionError: If the given sampler does not match the sampler API.

    See also:
        :class:`.Sampler` for the abstract base class that defines the sampler API.

    """

    # abstract base class

    assert isinstance(sampler, dimod.Sampler), "instantiated sampler must be a dimod Sampler object"

    # sample methods

    assert hasattr(sampler, 'sample'), "instantiated sampler must have a 'sample' method"
    assert callable(sampler.sample), "instantiated sampler must have a 'sample' method"

    assert hasattr(sampler, 'sample_ising'), "instantiated sampler must have a 'sample_ising' method"
    assert callable(sampler.sample_ising), "instantiated sampler must have a 'sample_ising' method"

    assert hasattr(sampler, 'sample_qubo'), "instantiated sampler must have a 'sample_qubo' method"
    assert callable(sampler.sample_qubo), "instantiated sampler must have a 'sample_qubo' method"

    # properties

    msg = "instantiated sampler must have a 'parameters' property, set to a Mapping"
    assert hasattr(sampler, 'parameters'), msg
    assert not callable(sampler.parameters), msg
    assert isinstance(sampler.parameters, abc.Mapping), msg

    msg = "instantiated sampler must have a 'properties' property, set to a Mapping"
    assert hasattr(sampler, 'properties'), msg
    assert not callable(sampler.properties), msg
    assert isinstance(sampler.properties, abc.Mapping), msg


def assert_composite_api(composed_sampler):
    """Assert that an instantiated composed sampler exposes correct composite
    properties and methods.

    Args:
        sampler (:obj:`.Composite`):
            User-made dimod composed sampler.

    Raises:
        AssertionError: If the given sampler does not match the composite API.

    See also:
        :class:`.Composite` for the abstract base class that defines the composite API.

        :obj:`~.assert_sampler_api` to assert that the composed sampler matches the sampler API.

    """

    assert isinstance(composed_sampler, dimod.Composite)

    # properties

    msg = "instantiated composed sampler must have a 'children' property, set to a list (or Sequence)"
    assert hasattr(composed_sampler, 'children'), msg
    assert isinstance(composed_sampler.children, abc.Sequence), msg

    msg = "instantiated composed sampler must have a 'child' property, set to one of sampler.children"
    assert hasattr(composed_sampler, 'child'), msg
    assert composed_sampler.child in composed_sampler.children, msg


def assert_structured_api(sampler):
    """Assert that an instantiated structured sampler exposes correct composite
    properties and methods.

    Args:
        sampler (:obj:`.Structured`):
            User-made dimod structured sampler.

    Raises:
        AssertionError: If the given sampler does not match the structured API.

    See also:
        :class:`.Structured` for the abstract base class that defines the structured API.

        :obj:`~.assert_sampler_api` to assert that the structured sampler matches the sampler API.

    """

    assert isinstance(sampler, dimod.Structured), "must be a Structured sampler"

    # properties

    msg = ("instantiated structured sampler must have an 'adjacency' property formatted as a dict "
           "where the keys are the nodes and the values are sets of all node adjacency to the key")
    assert hasattr(sampler, 'adjacency'), msg
    assert isinstance(sampler.adjacency, abc.Mapping), msg
    for u, neighborhood in sampler.adjacency.items():
        assert isinstance(neighborhood, abc.Set), msg
        for v in neighborhood:
            assert v in sampler.adjacency, msg
            assert u in sampler.adjacency[v], msg

    msg = "instantiated structured sampler must have a 'nodelist' property, set to a list"
    assert hasattr(sampler, 'nodelist'), msg
    assert isinstance(sampler.nodelist, abc.Sequence), msg
    for v in sampler.nodelist:
        assert v in sampler.adjacency, msg

    msg = "instantiated structured sampler must have a 'edge' property, set to a list of 2-lists/tuples"
    assert hasattr(sampler, 'edgelist'), msg
    assert isinstance(sampler.edgelist, abc.Sequence), msg
    for edge in sampler.edgelist:
        assert isinstance(edge, abc.Sequence), msg
        assert len(edge) == 2, msg

        u, v = edge
        assert v in sampler.adjacency, msg
        assert u in sampler.adjacency, msg
        assert u != v, msg


def assert_response_energies(response, bqm, precision=7):
    """Assert that each sample in the given response has the correct energy.

    Args:
        response (:obj:`.SampleSet`):
            Response as returned by a dimod sampler.

        bqm (:obj:`.BinaryQuadraticModel`):
            Binary quadratic model (BQM) used to generate the samples.

        precision (int, optional, default=7):
            Equality of energy is tested by calculating the difference between
            the `response`'s sample energy and that returned by BQM's
            :meth:`~.BinaryQuadraticModel.energy`, rounding to the closest
            multiple of 10 to the power of minus `precision`.

    Raises:
        AssertionError: If any of the samples in the response do not match their
        associated energy.

    See also:
        :func:`.assert_sampleset_energies`

    """
    return assert_sampleset_energies(response, bqm, precision)


def assert_sampleset_energies(sampleset, bqm, precision=7):
    """Assert that each sample in the given sample set has the correct energy.

    Args:
        sampleset (:obj:`.SampleSet`):
            Sample set as returned by a dimod sampler.

        bqm (:obj:`.BinaryQuadraticModel`/:obj:`.BinaryPolynomial`):
            The binary quadratic model (BQM) or binary polynomial used to generate the samples.

        precision (int, optional, default=7):
            Equality of energy is tested by calculating the difference between
            the `response`'s sample energy and that returned by BQM's
            :meth:`~.BinaryQuadraticModel.energy`, rounding to the closest
            multiple of 10 to the power of minus `precision`.

    Raises:
        AssertionError: If any of the samples in the sample set do not match
        their associated energy.

    Examples:

        >>> import dimod.testing
        ...
        >>> sampler = dimod.ExactSolver()
        >>> bqm = dimod.BinaryQuadraticModel.from_ising({}, {(0, 1): -1})
        >>> sampleset = sampler.sample(bqm)
        >>> dimod.testing.assert_response_energies(sampleset, bqm)

    """
    assert isinstance(sampleset, dimod.SampleSet), "expected sampleset to be a dimod SampleSet object"

    for sample, energy in sampleset.data(['sample', 'energy']):
        assert isinstance(sample, abc.Mapping), "'for sample in sampleset', each sample should be a Mapping"

        for v, value in sample.items():
            assert v in bqm.variables, 'sample contains a variable not in the given bqm'
            assert value in bqm.vartype.value, 'sample contains a value not of the correct type'

            for v in bqm.variables:
                assert v in sample, "bqm contains a variable not in sample"

        en = bqm.energy(sample)
        assert round(en - energy, precision) == 0, f"{en} is not almost equal to {energy}"


def assert_sampleset_energies_dqm(sampleset, dqm, precision=7):
    """Assert that each sample in the given sample set has the correct energy.

    Args:
        sampleset (:obj:`.SampleSet`):
            Sample set as returned by a dimod sampler.

        dqm (:obj:`.DiscreteQuadraticModel`):
            The discrete quadratic model (DQM) used to generate the samples.

        precision (int, optional, default=7):
            Equality of energy is tested by calculating the difference between
            the `response`'s sample energy and that returned by DQM's
            :meth:`~.DiscreteQuadraticModel.energy`, rounding to the closest
            multiple of 10 to the power of minus `precision`.

    Raises:
        AssertionError: If any of the samples in the sample set do not match
        their associated energy.

    """

    assert isinstance(sampleset, dimod.SampleSet), "expected sampleset to be a dimod SampleSet object"

    for sample, energy in sampleset.data(['sample', 'energy']):
        assert isinstance(sample, abc.Mapping), "'for sample in sampleset', each sample should be a Mapping"

        for v, value in sample.items():
            assert v in dqm.variables, 'sample contains a variable not in the given dqm'

            for v in dqm.variables:
                assert v in sample, "dqm contains a variable not in sample"

        assert round(dqm.energy(sample) - energy, precision) == 0
    
    
def assert_sampleset_energies_cqm(sampleset, cqm, precision=7):
    """Assert that each sample in the given sample set has the correct energy.

    Args:
        sampleset (:obj:`.SampleSet`):
            Sample set as returned by a dimod sampler.

        cqm (:obj:`.ConstrainedQuadraticModel`):
            The constrained quadratic model (CQM) used to generate the samples.

        precision (int, optional, default=7):
            Equality of energy is tested by calculating the difference between
            the `response`'s sample energy and that returned by CQM objective's
            :meth:`~.QuadraticModel.energy`, rounding to the closest
            multiple of 10 to the power of minus `precision`.

    Raises:
        AssertionError: If any of the samples in the sample set do not match
        their associated energy.

    """

    assert isinstance(sampleset, dimod.SampleSet), "expected sampleset to be a dimod SampleSet object"

    for sample, energy in sampleset.data(['sample', 'energy']):
        assert isinstance(sample, abc.Mapping), "'for sample in sampleset', each sample should be a Mapping"

        for v, value in sample.items():
            assert v in cqm.variables, 'sample contains a variable not in the given cqm'

            for v in cqm.variables:
                assert v in sample, "cqm contains a variable not in sample"

        assert round(cqm.objective.energy(sample) - energy, precision) == 0


def assert_bqm_almost_equal(actual, desired, places=7,
                            ignore_zero_interactions=False):
    """Test if two binary quadratic models have almost equal biases.

    Args:
        actual (:obj:`.BinaryQuadraticModel`):
            First binary quadratic model.

        desired (:obj:`.BinaryQuadraticModel`):
            Second binary quadratic model.

        places (int, optional, default=7):
            Bias equality is computed as :code:`round(b0 - b1, places) == 0`.

        ignore_zero_interactions (bool, optional, default=False):
            If true, interactions with 0 bias are ignored.

    """

    # vartype should match
    assert actual.vartype is desired.vartype, "unlike vartype"

    # variables should match
    variables_diff = set(actual.variables) ^ set(desired.variables)
    if variables_diff:
        v = variables_diff.pop()
        msg = "{!r} is not a shared variable".format(v)
        raise AssertionError(msg)

    # offset
    if round(actual.offset - desired.offset, places):
        msg = 'offsets {} != {}'.format(desired.offset, actual.offset)
        raise AssertionError(msg)

    # linear biases - we already checked variables
    for v, bias in desired.linear.items():
        if round(bias - actual.linear[v], places):
            msg = 'linear bias associated with {!r} does not match, {!r} != {!r}'
            raise AssertionError(msg.format(v, bias, actual.linear[v]))

    default = 0 if ignore_zero_interactions else None

    for inter, bias in actual.quadratic.items():
        other_bias = desired.quadratic.get(inter, default)
        if other_bias is None:
            raise AssertionError('{!r} is not a shared interaction'.format(inter))
        if round(bias - other_bias, places):
            msg = 'quadratic bias associated with {!r} does not match, {!r} != {!r}'
            raise AssertionError(msg.format(inter, bias, other_bias))
    for inter, bias in desired.quadratic.items():
        other_bias = actual.quadratic.get(inter, default)
        if other_bias is None:
            raise AssertionError('{!r} is not a shared interaction'.format(inter))
        if round(bias - other_bias, places):
            msg = 'quadratic bias associated with {!r} does not match, {!r} != {!r}'
            raise AssertionError(msg.format(inter, bias, other_bias))


def assert_consistent_bqm(bqm):
    """Test whether a BQM is self-consistent.

    This is useful when making new BQM subclasses. Asserts that all of the
    attributes are self-consistent.

    Args:
        bqm: A binary quadratic model.

    """
    # adjacency and linear are self-consistent
    for v in bqm.linear:
        assert v in bqm.adj
    for v in bqm.adj:
        assert v in bqm.linear

    # adjacency and quadratic are self-consistent
    for u, v in bqm.quadratic:
        assert v in bqm.linear
        assert v in bqm.adj
        assert u in bqm.adj[v]

        assert u in bqm.linear
        assert u in bqm.adj
        assert v in bqm.adj[u]

        assert bqm.adj[u][v] == bqm.quadratic[(u, v)]
        assert bqm.adj[u][v] == bqm.quadratic[(v, u)]
        assert bqm.adj[v][u] == bqm.adj[u][v]

    for u, v in bqm.quadratic:
        assert bqm.get_quadratic(u, v) == bqm.quadratic[(u, v)]
        assert bqm.get_quadratic(u, v) == bqm.quadratic[(v, u)]
        assert bqm.get_quadratic(v, u) == bqm.quadratic[(u, v)]
        assert bqm.get_quadratic(v, u) == bqm.quadratic[(v, u)]

    for u in bqm.adj:
        for v in bqm.adj[u]:
            assert (u, v) in bqm.quadratic
            assert (v, u) in bqm.quadratic

    assert len(bqm.quadratic) == bqm.num_interactions
    assert len(bqm.linear) == bqm.num_variables
    assert len(bqm.quadratic) == len(set(bqm.quadratic))
    assert len(bqm.variables) == len(bqm.linear)
    assert (bqm.num_variables, bqm.num_interactions) == bqm.shape



# distutils: language = c++
# cython: language_level=3

# Copyright 2021 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

cimport cython

from dimod.libcpp.quadratic_model cimport QuadraticModel as cppQuadraticModel
from dimod.libcpp.vartypes cimport Vartype as cppVartype


cdef class cyQM_template(cyQMBase):
    cdef cppQuadraticModel[bias_type, index_type]* cppqm

    cdef public int REAL_INTERACTIONS

    cdef Py_ssize_t _add_quadratic(self, index_type, index_type, bias_type) except -1
    cdef cppVartype cppvartype(self, object) except? cppVartype.SPIN
    cdef const cppQuadraticModel[bias_type, index_type]* data(self)



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
Solvers that calculate the energy of all possible samples.

Note:
    These samplers are designed for use in testing. Because they calculate
    energy for every possible sample, they are very slow.
"""
from itertools import product

import numpy as np

from dimod.binary.binary_quadratic_model import BinaryQuadraticModel
from dimod.constrained import ConstrainedQuadraticModel
from dimod.core.polysampler import PolySampler
from dimod.core.sampler import Sampler
from dimod.discrete.discrete_quadratic_model import DiscreteQuadraticModel
from dimod.higherorder.polynomial import BinaryPolynomial
from dimod.sampleset import SampleSet
from dimod.vartypes import Vartype

__all__ = ['ExactSolver', 'ExactPolySolver', 'ExactDQMSolver', 'ExactCQMSolver']


class ExactSolver(Sampler):
    """A simple exact solver for testing and debugging code using your local CPU.

    Notes:
        This solver becomes slow for problems with 18 or more
        variables.

    Examples:
        This example solves a two-variable Ising model.

        >>> h = {'a': -0.5, 'b': 1.0}
        >>> J = {('a', 'b'): -1.5}
        >>> sampleset = dimod.ExactSolver().sample_ising(h, J)
        >>> print(sampleset)
           a  b energy num_oc.
        0 -1 -1   -2.0       1
        2 +1 +1   -1.0       1
        1 +1 -1    0.0       1
        3 -1 +1    3.0       1
        ['SPIN', 4 rows, 4 samples, 2 variables]

    """
    properties = None
    parameters = None

    def __init__(self):
        self.properties = {}
        self.parameters = {}

    def sample(self, bqm: BinaryQuadraticModel, **kwargs) -> SampleSet:
        """Sample from a binary quadratic model.

        Args:
            bqm:
                Binary quadratic model to be sampled from.

        Returns:
            All possible solutions to the binary quadratic model.

        """
        kwargs = self.remove_unknown_kwargs(**kwargs)

        if not len(bqm.variables):
            return SampleSet.from_samples([], bqm.vartype, energy=[])

        samples = _graycode(bqm)

        if bqm.vartype is Vartype.SPIN:
            samples = 2*samples - 1

        return SampleSet.from_samples_bqm((samples, list(bqm.variables)), bqm)


class ExactPolySolver(PolySampler):
    """A simple exact polynomial solver for testing/debugging code on your CPU.

    Notes:
        This solver becomes slow for problems with 18 or more
        variables.

    Examples:
        This example solves a three-variable hising model.

        >>> h = {'a': -0.5, 'b': 1.0, 'c': 0.}
        >>> J = {('a', 'b'): -1.5, ('a', 'b', 'c'): -1.0}
        >>> sampleset = dimod.ExactPolySolver().sample_hising(h, J)
        >>> print(sampleset)                                # doctest: +SKIP
           a  b  c energy num_oc.
        1 -1 -1 +1   -3.0       1
        5 +1 +1 +1   -2.0       1
        0 -1 -1 -1   -1.0       1
        3 +1 -1 -1   -1.0       1
        4 +1 +1 -1    0.0       1
        2 +1 -1 +1    1.0       1
        7 -1 +1 -1    2.0       1
        6 -1 +1 +1    4.0       1
        ['SPIN', 8 rows, 8 samples, 3 variables]

        This example solves a three-variable binary polynomial

        >>> poly = dimod.BinaryPolynomial({('a',): 1.5, ('a', 'b'): -1, ('a', 'b', 'c'): 0.5}, 'SPIN')
        >>> sampleset = dimod.ExactPolySolver().sample_poly(poly)
        >>> sampleset.first.sample
        {'a': -1, 'b': -1, 'c': -1}

    """
    properties = None
    parameters = None

    def __init__(self):
        self.properties = {}
        self.parameters = {}

    def sample_poly(self, polynomial: BinaryPolynomial, **kwargs) -> SampleSet:
        """Sample from a binary polynomial.

        Args:
            polynomial:
                Binary polynomial to be sampled from.

        Returns:
            All possible solutions to the binary polynomial.

        """
        return ExactSolver().sample(polynomial, **kwargs)


class ExactDQMSolver():
    """A simple exact solver for testing and debugging code using your local CPU.

    Notes:
        This solver calculates the energy for every possible
        combination of variable cases. If variable :math:`i` has
        :math:`k_i` cases, this results in :math:`k_1 * k_2 * ... * k_n` cases,
        which grows exponentially for constant :math:`k_i` in the
        number of variables.

    """

    def __init__(self):
        self.properties = {}
        self.parameters = {}

    def sample_dqm(self, dqm: DiscreteQuadraticModel, **kwargs) -> SampleSet:
        """Sample from a discrete quadratic model.

        Args:
            dqm:
                Discrete quadratic model to be sampled from.

        Returns:
            All possible solutions to the discrete quadratic model.

        """
        Sampler.remove_unknown_kwargs(self, **kwargs)

        if not dqm.num_variables():
            return SampleSet.from_samples([], 'DISCRETE', energy=[])

        cases = _all_cases_dqm(dqm)
        energies = dqm.energies(cases)

        return SampleSet.from_samples(cases, 'DISCRETE', energies)


class ExactCQMSolver():
    """A simple exact solver for testing and debugging code using your local CPU.

    Notes:
        This solver calculates the energy for every possible
        combination of variable assignments. It becomes slow very quickly

    Examples:
        This example solves a CQM with 3 variables and 1 constraint.

        >>> from dimod import ConstrainedQuadraticModel, Binaries, ExactCQMSolver
        >>> cqm = ConstrainedQuadraticModel()
        >>> x, y, z = Binaries(['x', 'y', 'z'])
        >>> cqm.set_objective(x*y + 2*y*z)
        >>> cqm.add_constraint(x*y == 1, label='constraint_1')
        'constraint_1'
        >>> sampleset = ExactCQMSolver().sample_cqm(cqm)
        >>> print(sampleset)
          x y z energy num_oc. is_sat. is_fea.
        0 0 0 0    0.0       1 arra...   False
        1 0 1 0    0.0       1 arra...   False
        2 1 0 0    0.0       1 arra...   False
        4 0 0 1    0.0       1 arra...   False
        6 1 0 1    0.0       1 arra...   False
        3 1 1 0    1.0       1 arra...    True
        5 0 1 1    2.0       1 arra...   False
        7 1 1 1    3.0       1 arra...    True
        ['INTEGER', 8 rows, 8 samples, 3 variables]

    """
    def __init__(self):
        self.properties = {}
        self.parameters = {}

    def sample_cqm(self, cqm: ConstrainedQuadraticModel, *,
                   rtol: float = 1e-6,
                   atol: float = 1e-8,
                   **kwargs) -> SampleSet:
        """Sample from a constrained quadratic model.

        Args:
            cqm:
                Constrained quadratic model to be sampled from.
            rtol:
                Relative tolerance for constraint violation. The fraction of a
                constraint's right hand side which any violation must be smaller
                than for the constraint to be satisfied.
            atol:
                Absolute tolerance for constraint violations. A constant any
                violation must be smaller than for the constraint to be satisfied.

        Returns:
            A sampleset of all possible samples, with the fields ``is_feasible``
            and ``is_satisfied`` containing total and individual constraint
            violation information, respectively.

        """
        Sampler.remove_unknown_kwargs(self, **kwargs)

        if not len(cqm.variables):
            return SampleSet.from_samples([], 'INTEGER', energy=[])

        cases = _all_cases_cqm(cqm)

        return SampleSet.from_samples_cqm(cases, cqm, rtol=rtol, atol=atol, **kwargs)


def _graycode(bqm):
    """Get a numpy array containing all possible samples in a gray-code order"""
    # developer note: there are better/faster ways to do this, but because
    # we're limited in performance by the energy calculation, this is probably
    # more readable and easier.
    n = len(bqm.variables)
    ns = 1 << n
    samples = np.empty((ns, n), dtype=np.int8)

    samples[0, :] = 0

    for i in range(1, ns):
        v = (i & -i).bit_length() - 1  # the least significant set bit of i
        samples[i, :] = samples[i - 1, :]
        samples[i, v] = not samples[i - 1, v]

    return samples


def _all_cases_dqm(dqm):
    """Get a numpy array containing all possible samples as lists of integers"""
    # developer note: there may be better ways to do this, but because we're
    # limited in performance by the energy calculation, this is probably fine

    cases = [range(dqm.num_cases(v)) for v in dqm.variables]
    return np.array(np.meshgrid(*cases)).T.reshape(-1,dqm.num_variables()), list(dqm.variables)


def _all_cases_cqm(cqm):
    """Get a numpy array containing all possible samples as lists of valid values"""
    # developer note: the following is a catch all method, and it may be possible
    # to streamline or refactor. It is fine for its intended use

    var_list = list(cqm.variables)

    # Set aside logical discrete variables, and the Binary variables which make them up
    d_cases = [len(cqm.constraints[d].lhs.variables) for d in cqm.discrete]
    d_vars = [x for l in [list(cqm.constraints[d].lhs.variables) for d in cqm.discrete] for x in l]
    s = set(d_vars)
    var_list = [v for v in var_list if v not in s]

    # Construct all combinations of spin, integer, and binary variables not included in discrete
    cases = [_iterator_by_vartype(cqm, v) for v in var_list]
    c1 = np.array(np.meshgrid(*cases))
    if len(var_list):
        c1 = c1.T.reshape(-1,len(var_list))

    # Concatenate each discrete variable cases with each non-discrete variable cases
    combinations = []
    for indexes in product(*[range(d) for d in d_cases]):
        if not len(indexes):
            break

        l = []
        for i in range(len(indexes)):
            s = np.zeros(d_cases[i])
            s[indexes[i]] = 1
            l = np.concatenate([l, s])

        if len(c1):
            for row in c1:
                combinations.append(np.concatenate([l, row]))
        else:
            combinations.append(l)

    var_list = d_vars + var_list

    # If nothing was added to combinations, then there are no discrete variables
    if not len(combinations):
        combinations = c1

    return np.array(combinations), var_list


def _iterator_by_vartype(cqm, v):
    if cqm.vartype(v) is Vartype.BINARY:
        return range(2)
    if cqm.vartype(v) is Vartype.SPIN:
        return [-1, 1]
    if cqm.vartype(v) is Vartype.INTEGER:
        return range(int(cqm.lower_bound(v)), int(cqm.upper_bound(v)+1))
    raise ValueError("Only Binary, Spin, or Integer variables supported by ExactCQMSolver")



# Copyright 2021 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import abc
import collections.abc
import io
import operator

from collections.abc import ItemsView, MutableMapping
from typing import Any, Callable, Collection, Iterable, Iterator, Mapping, Optional, Tuple, Union

from dimod.typing import Bias, Variable

__all__ = ['Adjacency', 'Linear', 'Quadratic', 'QuadraticViewsMixin']


class TermsView:
    __slots__ = ['_model']

    def __init__(self, model: 'QuadraticViewsMixin'):
        self._model = model

    def __repr__(self):
        # let's just print the whole (potentially massive) thing for now, in
        # the future we'd like to do something a bit more clever (like hook
        # into dimod's Formatter)
        stream = io.StringIO()
        stream.write('{')
        last = len(self) - 1
        for i, (key, value) in enumerate(self.items()):
            stream.write(f'{key!r}: {value!r}')
            if i != last:
                stream.write(', ')
        stream.write('}')
        return stream.getvalue()


class Neighborhood(Mapping, TermsView):
    __slots__ = ['_var']

    def __init__(self, model: 'QuadraticViewsMixin', v: Variable):
        super().__init__(model)
        self._var = v

    def __getitem__(self, v: Variable) -> Bias:
        try:
            return self._model.get_quadratic(self._var, v)
        except ValueError as e:
            raise KeyError(*e.args)

    def __iter__(self) -> Iterator[Variable]:
        for v, _ in self._model.iter_neighborhood(self._var):
            yield v

    def __len__(self) -> int:
        return self._model.degree(self._var)

    def __setitem__(self, v: Variable, bias: Bias):
        self._model.set_quadratic(self._var, v, bias)

    def max(self, *, default: Optional[Bias] = None) -> Bias:
        """Return the maximum quadratic bias of the neighborhood"""
        try:
            return self._model.reduce_neighborhood(self._var, max)
        except TypeError as err:
            pass

        if default is None:
            raise ValueError("cannot find max of an empty sequence")

        return default

    def min(self, *, default: Optional[Bias] = None) -> Bias:
        """Return the minimum quadratic bias of the neighborhood"""
        try:
            return self._model.reduce_neighborhood(self._var, min)
        except TypeError as err:
            pass

        if default is None:
            raise ValueError("cannot find min of an empty sequence")

        return default

    def sum(self, start=0) -> Bias:
        """Return the sum of the quadratic biases of the neighborhood"""
        return self._model.reduce_neighborhood(self._var, operator.add, start)


class Adjacency(Mapping, TermsView):
    """Quadratic biases as a nested dict of dicts.

    Accessed like a dict of dicts, where the keys of the outer dict are all
    of the model's variables (e.g. `v`) and the values are the neighborhood of
    `v`. Each neighborhood is a dict where the keys are the neighbors of `v`
    and the values are their associated quadratic biases.
    """
    def __getitem__(self, v: Variable) -> Neighborhood:
        return Neighborhood(self._model, v)

    def __iter__(self) -> Iterator[Variable]:
        yield from self._model.variables

    def __len__(self) -> int:
        return len(self._model.variables)


class Linear(MutableMapping, TermsView):
    """Linear biases as a mapping.

    Accessed like a dict, where keys are the variables of the model and values
    are the linear biases.
    """
    def __delitem__(self, v: Variable):
        try:
            self._model.remove_variable(v)
        except ValueError:
            raise KeyError(repr(v))

    def __getitem__(self, v: Variable) -> Bias:
        try:
            return self._model.get_linear(v)
        except ValueError as e:
            raise KeyError(*e.args)

    def __iter__(self) -> Iterator[Variable]:
        yield from self._model.variables

    def __len__(self) -> int:
        return len(self._model.variables)

    def __setitem__(self, v: Variable, bias: Bias):
        self._model.set_linear(v, bias)

    def max(self, *, default: Optional[Bias] = None) -> Bias:
        """Return the maximum linear bias."""
        try:
            return self._model.reduce_linear(max)
        except TypeError:
            pass

        if default is None:
            raise ValueError("cannot find max of an empty sequence")

        return default

    def min(self, *, default: Optional[Bias] = None) -> Bias:
        """Return the minimum linear bias."""
        try:
            return self._model.reduce_linear(min)
        except TypeError:
            pass

        if default is None:
            raise ValueError("cannot find min of an empty sequence")

        return default

    def sum(self, start=0):
        """Return the sum of the linear biases."""
        return self._model.reduce_linear(operator.add, start)


class QuadraticItemsView(ItemsView):
    # speed up iteration
    def __iter__(self) -> Iterator[Tuple[Tuple[Variable, Variable], Bias]]:
        for u, v, bias in self._mapping._model.iter_quadratic():
            yield (u, v), bias


class Quadratic(MutableMapping, TermsView):
    """Quadratic biases as a flat mapping.

    Accessed like a dict, where keys are 2-tuples of variables, which represent
    an interaction and values are the quadratic biases.
    """
    def __delitem__(self, uv: Tuple[Variable, Variable]):
        try:
            self._model.remove_interaction(*uv)
        except ValueError:
            raise KeyError(repr(uv))

    def __eq__(self, other):
        if not isinstance(other, Mapping):
            return NotImplemented

        try:
            return (len(self) == len(other) and
                    all(self[key] == value for key, value in other.items()))
        except KeyError:
            return False

    def __getitem__(self, uv: Tuple[Variable, Variable]) -> Bias:
        try:
            return self._model.get_quadratic(*uv)
        except ValueError as e:
            raise KeyError(*e.args)

    def __iter__(self) -> Iterator[Tuple[Variable, Variable]]:
        for u, v, _ in self._model.iter_quadratic():
            yield u, v

    def __len__(self) -> int:
        return self._model.num_interactions

    def __setitem__(self, uv: Tuple[Variable, Variable], bias: Bias):
        self._model.set_quadratic(*uv, bias)

    def items(self) -> ItemsView:
        return QuadraticItemsView(self)

    def max(self, *, default: Optional[Bias] = None) -> Bias:
        """Return the maximum quadratic bias."""
        try:
            return self._model.reduce_quadratic(max)
        except TypeError:
            pass

        if default is None:
            raise ValueError("cannot find max of an empty sequence")

        return default

    def min(self, *, default: Optional[Bias] = None) -> Bias:
        """Return the minimum quadratic bias."""
        try:
            return self._model.reduce_quadratic(min)
        except TypeError:
            pass

        if default is None:
            raise ValueError("cannot find min of an empty sequence")

        return default

    def sum(self, start=0):
        """Return the sum of the quadratic biases."""
        return self._model.reduce_quadratic(operator.add, start)


class QuadraticViewsMixin(abc.ABC):

    @property
    def adj(self) -> Adjacency:
        """Adjacency structure as a nested mapping of mappings.

        Accessed like a dict of dicts, where the keys of the outer dict are all
        of the model's variables (e.g. ``v``) and the values are the neighborhood
        of ``v``. Each neighborhood is a dict where the keys are the neighbors of
        ``v`` and the values are their associated quadratic biases.

        Examples:
            >>> from dimod import QuadraticModel, Binary, Integer
            >>> qm = QuadraticModel()
            >>> qm.add_variables_from('BINARY', ['x', 'y'])
            >>> qm.add_variables_from('INTEGER', ['i', 'j'])
            >>> qm.add_quadratic('i', 'j', 2)
            >>> qm.add_quadratic('x', 'i', -1)
            >>> qm.adj
            {'x': {'i': -1.0}, 'y': {}, 'i': {'x': -1.0, 'j': 2.0}, 'j': {'i': 2.0}}

        """
        # we could use cached property but this is way simpler and doesn't
        # break the docs
        try:
            return self._adj  # type: ignore[has-type]
        except AttributeError:
            pass
        self._adj = adj = Adjacency(self)
        return adj

    @property
    def linear(self) -> Linear:
        """Linear biases as a mapping.

        Accessed like a dict, where keys are the variables of the model and
        values are the linear biases.

        Examples:
            >>> from dimod import QuadraticModel, Binary, Integer
            >>> qm = QuadraticModel()
            >>> qm.add_variables_from('BINARY', ['x', 'y'])
            >>> qm.add_variables_from('INTEGER', ['i', 'j'])
            >>> qm.add_linear('x', 0.5)
            >>> qm.add_linear('i', -2)
            >>> qm.linear
            {'x': 0.5, 'y': 0.0, 'i': -2.0, 'j': 0.0}

        """
        # we could use cached property but this is way simpler and doesn't
        # break the docs
        try:
            return self._linear  # type: ignore[has-type]
        except AttributeError:
            pass
        self._linear = linear = Linear(self)
        return linear

    @property
    def quadratic(self) -> Quadratic:
        """Quadratic biases as a flat mapping.

        Accessed like a dict, where keys are 2-tuples of variables, which
        represent an interaction and values are the quadratic biases.

        Examples:
            >>> from dimod import QuadraticModel, Binary, Integer
            >>> qm = QuadraticModel()
            >>> qm.add_variables_from('BINARY', ['x', 'y'])
            >>> qm.add_variables_from('INTEGER', ['i', 'j'])
            >>> qm.add_quadratic('i', 'j', 2)
            >>> qm.add_quadratic('x', 'i', -1)
            >>> qm.quadratic
            {('i', 'x'): -1.0, ('j', 'i'): 2.0}
        """
        # we could use cached property but this is way simpler and doesn't
        # break the docs
        try:
            return self._quadratic  # type: ignore[has-type]
        except AttributeError:
            pass
        self._quadratic = quadratic = Quadratic(self)
        return quadratic

    @property
    @abc.abstractmethod
    def num_interactions(self) -> int:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def offset(self) -> Bias:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def variables(self) -> Collection[Variable]:
        raise NotImplementedError

    @abc.abstractmethod
    def add_linear(self, v: Variable, bias: Bias):
        raise NotImplementedError

    @abc.abstractmethod
    def degree(self, v: Variable) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def get_linear(self, v: Variable) -> Bias:
        raise NotImplementedError

    @abc.abstractmethod
    def get_quadratic(self, u: Variable, v: Variable) -> Bias:
        raise NotImplementedError

    @abc.abstractmethod
    def iter_neighborhood(self, v: Variable) -> Iterator[Tuple[Variable, Bias]]:
        raise NotImplementedError

    @abc.abstractmethod
    def iter_quadratic(self) -> Iterator[Tuple[Variable, Variable, Bias]]:
        raise NotImplementedError

    @abc.abstractmethod
    def reduce_linear(self, function: Callable, initializer: Optional[Bias] = None) -> Any:
        raise NotImplementedError

    @abc.abstractmethod
    def reduce_neighborhood(self, v: Variable, function: Callable,
                            initializer: Optional[Bias] = None) -> Any:
        raise NotImplementedError

    @abc.abstractmethod
    def reduce_quadratic(self, function: Callable, initializer: Optional[Bias] = None) -> Any:
        raise NotImplementedError

    @abc.abstractmethod
    def remove_interaction(self, u: Variable, v: Variable):
        raise NotImplementedError

    @abc.abstractmethod
    def remove_variable(self, v: Variable) -> Variable:
        raise NotImplementedError

    @abc.abstractmethod
    def set_linear(self, v: Variable, bias: Bias):
        raise NotImplementedError

    @abc.abstractmethod
    def set_quadratic(self, u: Variable, v: Variable, bias: Bias):
        raise NotImplementedError

    def add_linear_from(self, linear: Union[Iterable, Mapping]):
        """Add variables and linear biases to a binary quadratic model.

        Args:
            linear:
                Variables and their associated linear biases, as either a dict of
                form ``{v: bias, ...}`` or an iterable of ``(v, bias)`` pairs,
                where ``v`` is a variable and ``bias`` is its associated linear
                bias.

        """
        if isinstance(linear, collections.abc.Mapping):
            iterator = linear.items()
        elif isinstance(linear, collections.abc.Iterable):
            iterator = linear
        else:
            raise TypeError(
                "expected 'linear' to be a dict or an iterable of 2-tuples.")

        for v, bias in iterator:
            self.add_linear(v, bias)

    add_variables_from = add_linear_from
    """Alias for :meth:`add_linear_from`."""

    def fix_variable(self, v: Variable, value: float):
        """Remove a variable by fixing its value.

        Args:
            v: Variable to be fixed.

            value: Value assigned to the variable. Values should generally
                match the :class:`.Vartype` of the variable, but do not have
                to.

        Raises:
            ValueError: If ``v`` is not a variable in the model.

        """
        add_linear = self.add_linear
        for u, bias in self.iter_neighborhood(v):
            add_linear(u, value*bias)

        self.offset += value*self.get_linear(v)
        self.remove_variable(v)

    def fix_variables(self,
                      fixed: Union[Mapping[Variable, float], Iterable[Tuple[Variable, float]]]):
        """Fix the value of the variables and remove them.

        Args:
            fixed: A dictionary or an iterable of 2-tuples of variable
                assignments.

        """
        if isinstance(fixed, Mapping):
            fixed = fixed.items()

        fix_variable = self.fix_variable
        for v, val in fixed:
            fix_variable(v, val)

    def iter_linear(self) -> Iterator[Tuple[Variable, Bias]]:
        """Iterate over the variables and their biases."""
        get = self.get_linear
        for v in self.variables:
            yield v, get(v)

    def to_polystring(self, encoder: Optional[Callable[[Variable], str]] = None) -> str:
        """Return a string representing the model as a polynomial.

        Args:
            encoder: A function mapping variables to a string. By default
                string variables are mapped directly whereas all other types
                are mapped to a string :code:`f"v{variable!r}"`.

        Returns:
            A string representing the binary quadratic model.

        Examples:

            >>> x, y, z = dimod.Binaries(['x', 'y', 'z'])
            >>> (2*x + 3*y*z + 6).to_polystring()
            '6 + 2*x + 3*y*z'

        """

        if encoder is None:
            def encoder(v: Variable) -> str:
                return v if isinstance(v, str) else f"v{v!r}"

        # developer note: we use floats everywhere because they have some nice
        # methods and because for this method we're not too worried about
        # performance

        def neg(bias: float) -> bool: return bias < 0

        def string(bias: float) -> str:
            return repr(abs(int(bias))) if bias.is_integer() else repr(abs(bias))

        def coefficient(bias: float) -> str:
            return '' if abs(bias) == 1 else f"{string(bias)}*"

        # linear variables that have a positive bias or are not going to be
        # included implicitly in quadratic
        linear = ((v, float(bias)) for v, bias in self.iter_linear()
                  if bias or not any(bias for _, bias in self.iter_neighborhood(v)))

        # non-zero quadratic biases
        quadratic = ((u, v, float(bias)) for u, v, bias in self.iter_quadratic() if bias)

        # offset
        offset = float(self.offset)

        sio = io.StringIO()

        # the first element is special, since we put the sign adjacent to it so
        # let's handle that case
        if offset or not len(self.variables):
            if neg(offset):
                sio.write('-')
            sio.write(string(offset))

        else:
            # the first element can come from quadratic or linear

            try:
                v, bias = next(linear)
            except StopIteration:
                # we are guaranteed that this exists
                u, v, bias = next(quadratic)
                if neg(bias):
                    sio.write('-')
                sio.write(f"{coefficient(bias)}{encoder(v)}*{encoder(u)}")
            else:
                # there is a linear bias
                if neg(bias):
                    sio.write('-')
                sio.write(f'{coefficient(bias)}{encoder(v)}')

        for v, bias in linear:
            sio.write(f" {'-' if neg(bias) else '+'} {coefficient(bias)}{encoder(v)}")

        for u, v, bias in quadratic:
            sio.write(f" {'-' if neg(bias) else '+'} {coefficient(bias)}{encoder(v)}*{encoder(u)}")

        sio.seek(0)
        return sio.read()



# Copyright 2019 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
A composite that truncates the returned :obj:`dimod.SampleSet` based on options
specified by the user.
"""

import numpy as np

from dimod.core.composite import ComposedSampler

__all__ = 'TruncateComposite',


class TruncateComposite(ComposedSampler):
    """Composite to truncate the returned sample set.

    Inherits from :class:`dimod.ComposedSampler`.

    Post-processing can be expensive and sometimes you might want to only
    handle the lowest-energy samples. This composite layer allows you to
    pre-select the samples within a multi-composite pipeline.

    Args:
        child_sampler (:obj:`dimod.Sampler`):
            A dimod sampler.

        n (int):
            Maximum number of rows in the returned sample set.

        sorted_by (str/None, optional, default='energy'):
            Selects the record field used to sort the samples before
            truncating. Note that sample order is maintained in the
            underlying array.

        aggregate (bool, optional, default=False):
            If True, aggregates the samples before truncating and sets the value
            of the ``num_occurrences`` field in the returned :class:`~dimod.SampleSet`
            to the number of accumulated samples for each occurrence.

    Examples:
        >>> sampler = dimod.TruncateComposite(dimod.RandomSampler(), n=2, aggregate=True)
        >>> bqm = dimod.BinaryQuadraticModel.from_ising({"a": 1, "b": 2}, {("a", "b"): -1})
        >>> sampleset = sampler.sample(bqm, num_reads=100)
        >>> print(sampleset)                                 # doctest:+SKIP
           a  b energy num_oc.
        0 -1 -1   -4.0      16
        1 +1 -1    0.0      22
        ['SPIN', 2 rows, 38 samples, 2 variables]

    """

    def __init__(self, child_sampler, n, sorted_by='energy', aggregate=False):

        if n < 1:
            raise ValueError('n should be a positive integer, recived {}'.format(n))

        self._children = [child_sampler]
        self._truncate_kwargs = dict(n=n, sorted_by=sorted_by)
        self._aggregate = aggregate

    @property
    def children(self):
        return self._children

    @property
    def parameters(self):
        return self.child.parameters.copy()

    @property
    def properties(self):
        return {'child_properties': self.child.properties.copy()}

    def sample(self, bqm, **kwargs):
        """Sample from the binary quadratic model and truncate returned sample set.

        Args:
            bqm (:obj:`dimod.BinaryQuadraticModel`):
                Binary quadratic model to be sampled from.

            **kwargs:
                Parameters for the sampling method, specified by the child
                sampler.

        Returns:
            :obj:`dimod.SampleSet`

        """
        tkw = self._truncate_kwargs
        if self._aggregate:
            return self.child.sample(bqm, **kwargs).aggregate().truncate(**tkw)
        else:
            return self.child.sample(bqm, **kwargs).truncate(**tkw)



# distutils: include_dirs = dimod/include/

# Copyright 2022 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from libcpp.utility cimport pair

from dimod.libcpp.abc cimport QuadraticModelBase
from dimod.libcpp.vartypes cimport Vartype

__all__ = ['QuadraticModel']


cdef extern from "dimod/quadratic_model.h" namespace "dimod" nogil:
    cdef cppclass QuadraticModel[Bias, Index](QuadraticModelBase[Bias, Index]):
        ctypedef Bias bias_type
        ctypedef Index index_type
        ctypedef size_t size_type

        pair[const_neighborhood_iterator, const_neighborhood_iterator] neighborhood(index_type)  # deprecated
        pair[const_neighborhood_iterator, const_neighborhood_iterator] neighborhood(index_type, index_type)  # deprecated

        # Methods/attributes specific to the QM
        index_type add_variable(Vartype)
        index_type add_variable(Vartype, bias_type, bias_type)
        index_type add_variables(Vartype, index_type)
        index_type add_variables(Vartype, index_type, bias_type bias_type)
        void change_vartype(Vartype, index_type) except+
        void resize(index_type) except+
        void resize(index_type, Vartype) except+
        void resize(index_type, Vartype, bias_type, bias_type)
        void set_lower_bound(index_type, bias_type)
        void set_upper_bound(index_type, bias_type)
        void set_vartype(index_type, Vartype)



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
Composites that convert binary quadratic model samplers into polynomial samplers
or that work with binary polynomials.

Higher-order composites implement three sampling methods (similar to
:class:`.Sampler`):

* :meth:`.PolySampler.sample_poly`
* :meth:`.PolySampler.sample_hising`
* :meth:`.PolySampler.sample_hubo`

"""
import warnings

from collections import defaultdict

import numpy as np

from dimod.core.polysampler import ComposedPolySampler, PolySampler
from dimod.higherorder.polynomial import BinaryPolynomial
from dimod.higherorder.utils import make_quadratic, poly_energies
from dimod.sampleset import SampleSet, append_variables



__all__ = ['HigherOrderComposite',
           'PolyScaleComposite',
           'PolyTruncateComposite',
           'PolyFixedVariableComposite',
           ]


class HigherOrderComposite(ComposedPolySampler):
    """Convert a binary quadratic model sampler to a binary polynomial sampler.

    Energies of the returned samples do not include the penalties.

    Args:
        sampler (:obj:`dimod.Sampler`):
            A dimod sampler

    Example:
        This example uses :class:`.HigherOrderComposite` to instantiate a
        composed sampler that submits a simple Ising problem to a sampler.
        The composed sampler creates a binary quadratic model (BQM) from a
        higher order problem.

        >>> sampler = dimod.HigherOrderComposite(dimod.ExactSolver())
        >>> h = {0: -0.5, 1: -0.3, 2: -0.8}
        >>> J = {(0, 1, 2): -1.7}
        >>> sampleset = sampler.sample_hising(h, J, discard_unsatisfied=True)
        >>> set(sampleset.first.sample.values()) == {1}
        True

        """

    def __init__(self, child_sampler):
        self._children = [child_sampler]

    @property
    def children(self):
        """A list containing the wrapped sampler."""
        return self._children

    @property
    def parameters(self):
        param = self.child.parameters.copy()
        param['penalty_strength'] = []
        param['discard_unsatisfied'] = []
        param['keep_penalty_variables'] = []
        return param

    @property
    def properties(self):
        return {'child_properties': self.child.properties.copy()}

    def sample_poly(self, poly, penalty_strength=1.0,
                    keep_penalty_variables=False,
                    discard_unsatisfied=False, **parameters):
        """Sample from the given binary polynomial.

        Introduces penalties to reduce the given higher-order binary polynomial
        to a quadratic problem and sends it to its child sampler.

        Args:
            poly (:obj:`.BinaryPolynomial`):
                A binary polynomial.

            penalty_strength (float, optional):
                Strength of the reduction constraint. Insufficient strength can
                result in the binary quadratic model not having the same
                minimization as the polynomial.

            keep_penalty_variables (bool, optional, default=True):
                Setting to False removes the variables used for penalty from
                the samples.

            discard_unsatisfied (bool, optional, default=False):
                Setting to True discards samples that do not satisfy the penalty
                conditions.

            initial_state (dict, optional):
                Only accepted when the child sampler accepts an initial state.
                The initial state is given in terms of the variables in
                the binary polynomial. The corresponding initial values are
                populated for use by the child sampler.

            **parameters:
                Parameters for the sampling method, specified by the child
                sampler.

        Returns:
            :obj:`dimod.SampleSet`

        """

        bqm = make_quadratic(poly, penalty_strength, vartype=poly.vartype)

        if 'initial_state' in parameters:
            initial_state = expand_initial_state(bqm,
                                                 parameters['initial_state'])
            parameters['initial_state'] = initial_state

        response = self.child.sample(bqm, **parameters)

        return polymorph_response(response, poly, bqm,
                                  penalty_strength=penalty_strength,
                                  keep_penalty_variables=keep_penalty_variables,
                                  discard_unsatisfied=discard_unsatisfied)


def expand_initial_state(bqm, initial_state):
    """Determine values for the initial state.

    Used on a binary quadratic model generated from a higher order polynomial.

    Args:
        bqm (:obj:`.BinaryQuadraticModel`): a binary quadratic model (BQM)
        object that contains its reduction info.

        initial_state (dict):
            An initial state for the higher order polynomial that generated the
            BQM.

    Returns:
        dict: A fully specified initial state.

    """
    # Developer note: this function relies heavily on assumptions about the
    # existance and structure of bqm.info['reduction']. We should consider
    # changing the way that the reduction information is passed.
    if not bqm.info['reduction']:
        return initial_state  # saves making a copy

    initial_state = dict(initial_state)  # so we can edit it in-place

    for (u, v), changes in bqm.info['reduction'].items():

        uv = changes['product']
        initial_state[uv] = initial_state[u] * initial_state[v]

        if 'auxiliary' in changes:
            # need to figure out the minimization from the initial_state
            aux = changes['auxiliary']

            en = (initial_state[u] * bqm.adj[aux].get(u, 0) +
                  initial_state[v] * bqm.adj[aux].get(v, 0) +
                  initial_state[uv] * bqm.adj[aux].get(uv, 0))

            initial_state[aux] = min(bqm.vartype.value, key=lambda val: en*val)

    return initial_state


def penalty_satisfaction(response, bqm):
    """Create a penalty satisfaction list.

    Given a sample set and a binary quadratic model (BQM) object, creates a
    binary list informing whether the penalties introduced during degree
    reduction are satisfied for each sample in the sample set.

    Args:
        response (:obj:`.SampleSet`): Samples corresponding to provided BQM

        bqm (:obj:`.BinaryQuadraticModel`): a BQM object that contains
            its reduction info.

    Returns:
        :obj:`numpy.ndarray`: a binary array of penalty satisfaction information

    """
    record = response.record
    label_to_idx = response.variables.index

    if len(bqm.info['reduction']) == 0:
        return np.array([1] * len(record.sample))

    penalty_vector = np.prod([record.sample[:, label_to_idx(qi)] *
                              record.sample[:, label_to_idx(qj)]
                              == record.sample[:,
                                 label_to_idx(valdict['product'])]
                              for (qi, qj), valdict in
                              bqm.info['reduction'].items()], axis=0)
    return penalty_vector


def polymorph_response(response, poly, bqm,
                       penalty_strength=None,
                       keep_penalty_variables=True,
                       discard_unsatisfied=False):
    """Transform the sample set for the higher order problem.

    Given a response of a penalized HUBO, this function creates a new sample set
    object, taking into account penalty information and calculates the
    energies of samples for the higherorder problem.

    Args:
        response (:obj:`.SampleSet`): response for a penalized hubo.

        poly (:obj:`.BinaryPolynomial`):
            A binary polynomial.

        bqm (:obj:`dimod.BinaryQuadraticModel`): Binary quadratic model of the
            reduced problem.

        penalty_strength (float, optional, default=None):
            If provided, added to the info field of the
            returned :obj:`dimod.SampleSet`.

        keep_penalty_variables (bool, optional, default=True):
            Setting to False removes variables used for penalty from the samples.

        discard_unsatisfied (bool, optional, default=False):
            Setting to True discards samples that do not satisfy the penalty
            conditions.

    Returns:
        (:obj:`.SampleSet'): A SampleSet object that has additional penalty
            information. The energies of samples are calculated for the HUBO
            ignoring the penalty variables.

    """
    record = response.record
    penalty_vector = penalty_satisfaction(response, bqm)
    original_variables = bqm.variables

    if discard_unsatisfied:
        samples_to_keep = list(map(bool, list(penalty_vector)))
        penalty_vector = np.array([True] * np.sum(samples_to_keep))
    else:
        samples_to_keep = list(map(bool, [1] * len(record.sample)))

    samples = record.sample[samples_to_keep]
    energy_vector = poly.energies((samples, response.variables))

    if not keep_penalty_variables:
        original_variables = poly.variables
        idxs = [response.variables.index(v) for v in original_variables]
        samples = np.asarray(samples[:, idxs])

    num_samples, num_variables = np.shape(samples)

    datatypes = [('sample', np.dtype(np.int8), (num_variables,)),
                 ('energy', energy_vector.dtype),
                 ('penalty_satisfaction',
                  penalty_vector.dtype)]
    datatypes.extend((name, record[name].dtype, record[name].shape[1:])
                     for name in record.dtype.names if
                     name not in {'sample',
                                  'energy'})

    data = np.rec.array(np.empty(num_samples, dtype=datatypes))
    data.sample = samples
    data.energy = energy_vector
    for name in record.dtype.names:
        if name not in {'sample', 'energy'}:
            data[name] = record[name][samples_to_keep]

    data['penalty_satisfaction'] = penalty_vector
    response.info['reduction'] = bqm.info['reduction']
    if penalty_strength is not None:
        response.info['penalty_strength'] = penalty_strength
    return SampleSet(data, original_variables, response.info,
                     response.vartype)


class PolyScaleComposite(ComposedPolySampler):
    """Composite to scale biases of a binary polynomial.

    Args:
        child (:obj:`.PolySampler`):
            A binary polynomial sampler.

    Examples:

       >>> linear = {'a': -4.0, 'b': -4.0}
       >>> quadratic = {('a', 'b'): 3.2, ('a', 'b', 'c'): 1}
       >>> sampler = dimod.PolyScaleComposite(dimod.HigherOrderComposite(dimod.ExactSolver()))
       >>> response = sampler.sample_hising(linear, quadratic, scalar=0.5,
       ...                ignored_terms=[('a','b')])

    """

    def __init__(self, child):
        if not isinstance(child, PolySampler):
            raise TypeError("Child sampler must be a PolySampler")
        self._children = [child]

    @property
    def children(self):
        """The child sampler in a list"""
        return self._children

    @property
    def parameters(self):
        param = self.child.parameters.copy()
        param.update({'scalar': [],
                      'bias_range': [],
                      'poly_range': [],
                      'ignored_terms': [],
                      })
        return param

    @property
    def properties(self):
        return {'child_properties': self.child.properties.copy()}

    def sample_poly(self, poly, scalar=None, bias_range=1, poly_range=None,
                    ignored_terms=None, **parameters):
        """Scale and sample from the given binary polynomial.

        If `scalar` is not given, problem is scaled based on bias and polynomial
        ranges. See :meth:`.BinaryPolynomial.scale` and
        :meth:`.BinaryPolynomial.normalize`

        Args:
            poly (obj:`.BinaryPolynomial`): A binary polynomial.

            scalar (number, optional):
                Value by which to scale the energy range of the binary polynomial.

            bias_range (number/pair, optional, default=1):
                Value/range by which to normalize all the biases, or if
                `poly_range` is provided, just the linear biases.

            poly_range (number/pair, optional):
                Value/range by which to normalize higher-order biases.

            ignored_terms (iterable, optional):
                Biases associated with these terms are not scaled.

            **parameters:
                Other parameters for the sampling method, specified by
                the child sampler.

        """

        if ignored_terms is None:
            ignored_terms = set()
        else:
            ignored_terms = {frozenset(term) for term in ignored_terms}

        # scale and normalize happen in-place so we need to make a copy
        original, poly = poly, poly.copy()

        if scalar is not None:
            poly.scale(scalar, ignored_terms=ignored_terms)
        else:
            poly.normalize(bias_range=bias_range, poly_range=poly_range,
                           ignored_terms=ignored_terms)

            # we need to know how much we scaled by, which we can do by looking
            # at the biases
            try:
                v = next(v for v, bias in original.items()
                         if bias and v not in ignored_terms)
            except StopIteration:
                # nothing to scale
                scalar = 1
            else:
                scalar = poly[v] / original[v]

        sampleset = self.child.sample_poly(poly, **parameters)

        if ignored_terms:
            # we need to recalculate the energy
            sampleset.record.energy = original.energies((sampleset.record.sample,
                                                         sampleset.variables))
        else:
            sampleset.record.energy /= scalar

        return sampleset


class PolyTruncateComposite(ComposedPolySampler):
    """Composite that truncates returned samples.

    Post-processing is expensive and sometimes one might want to only
    treat the lowest-energy samples. This composite layer allows one to
    pre-select the samples within a multi-composite pipeline.

    Args:
        child_sampler (:obj:`dimod.PolySampler`):
            A dimod binary polynomial sampler.

        n (int):
            Maximum number of rows in the returned sample set.

        sorted_by (str/None, optional, default='energy'):
            Selects the record field used to sort the samples before
            truncating. Note that sample order is maintained in the
            underlying array.

        aggregate (bool, optional, default=False):
            If True, aggregate the samples before truncating.

    Note:
        If `aggregate` is True, :attr:`.SampleSet.record.num_occurrences` are
        accumulated but no other fields are.

    """
    def __init__(self, child_sampler, n, sorted_by='energy', aggregate=False):

        if n < 1:
            raise ValueError('n should be a positive integer, recived {}'.format(n))

        self._children = [child_sampler]
        self._truncate_kwargs = dict(n=n, sorted_by=sorted_by)
        self._aggregate = aggregate

    @property
    def children(self):
        return self._children

    @property
    def parameters(self):
        return self.child.parameters.copy()

    @property
    def properties(self):
        return {'child_properties': self.child.properties.copy()}

    def sample_poly(self, poly, **kwargs):
        """Sample from the binary polynomial and truncate output.

        Args:
            poly (obj:`.BinaryPolynomial`): A binary polynomial.

            **kwargs:
                Parameters for the sampling method, specified by the child
                sampler.

        Returns:
            :obj:`dimod.SampleSet`

        """
        tkw = self._truncate_kwargs
        if self._aggregate:
            return self.child.sample_poly(poly, **kwargs).aggregate().truncate(**tkw)
        else:
            return self.child.sample_poly(poly, **kwargs).truncate(**tkw)


class PolyFixedVariableComposite(ComposedPolySampler):
    """Composite that fixes variables of a problem.

    Fixes variables of a binary polynomial and modifies linear and k-local terms
    accordingly. Returned samples include the fixed variable.

    Args:
       sampler (:obj:`dimod.PolySampler`):
            A dimod polynomial sampler.

    Examples:
       This example uses :class:`.PolyFixedVariableComposite` to instantiate a
       composed sampler that submits a simple high-order Ising problem to a sampler.
       The composed sampler fixes a variable and modifies linear and k-local terms
       biases.

       >>> h = {1: -1.3, 2: 1.2, 3: -3.4, 4: -0.5}
       >>> J = {(1, 4): -0.6, (1, 2, 3): 0.2, (1, 2, 3, 4): -0.1}
       >>> poly = dimod.BinaryPolynomial.from_hising(h, J, offset=0)
       >>> sampler = dimod.PolyFixedVariableComposite(dimod.ExactPolySolver())
       >>> sampleset = sampler.sample_poly(poly, fixed_variables={3: -1, 4: 1})

    """

    def __init__(self, child_sampler):
        self._children = [child_sampler]

    @property
    def children(self):
        return self._children

    @property
    def parameters(self):
        params = self.child.parameters.copy()
        params['fixed_variables'] = []
        return params

    @property
    def properties(self):
        return {'child_properties': self.child.properties.copy()}

    def sample_poly(self, poly, fixed_variables=None, **parameters):
        """Sample from the provided binary quadratic model.

        Args:
            poly (:obj:`dimod.BinaryPolynomial`):
                Binary polynomial model to be sampled from.

            fixed_variables (dict):
                A dictionary of variable assignments.

            **parameters:
                Parameters for the sampling method, specified by the child sampler.

        Returns:
            :obj:`dimod.SampleSet`

        """

        child = self.child
        if fixed_variables is None:
            sampleset = child.sample_poly(poly, **parameters)
            return sampleset
        else:
            poly_copy = fix_variables(poly, fixed_variables)
            sampleset = child.sample_poly(poly_copy, **parameters)
            if len(sampleset):
                return append_variables(sampleset, fixed_variables)
            elif fixed_variables:
                return type(sampleset).from_samples_bqm(fixed_variables, bqm=poly)
            else:
                return sampleset


def fix_variables(poly, fixed_variables):
    if () in poly.keys():
        offset = poly[()]
    else:
        offset = 0.0
    poly_copy = defaultdict(float)
    for k, v in poly.items():
        k = set(k)
        for var, value in fixed_variables.items():
            if var in k:
                k -= {var}
                v *= value
        k = frozenset(k)
        if len(k) > 0:
            poly_copy[k] += v
        else:
            offset += v
    poly_copy[()] = offset
    return BinaryPolynomial(poly_copy, poly.vartype)



# Copyright 2022 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from __future__ import annotations

import collections.abc
import itertools
import typing

import numpy as np

import dimod  # for typing

from dimod.binary import BinaryQuadraticModel
from dimod.vartypes import Vartype


__all__ = ["random_nae3sat", "random_2in4sat"]


def _kmcsat_interactions(num_variables: int, k: int, num_clauses: int,
                         *,
                         plant_solution: bool = False,
                         seed: typing.Union[None, int, np.random.Generator] = None,
                         ) -> typing.Iterator[typing.Tuple[int, int, int]]:
    rng = np.random.default_rng(seed)

    # Use of for and while loops is for clarity, optimizations are possible.
    for _ in range(num_clauses):
        # randomly select the variables
        variables = rng.choice(num_variables, k, replace=False)

        # randomly assign the negations
        signs = 2 * rng.integers(0, 1, endpoint=True, size=k) - 1
        while plant_solution and abs(sum(signs))>1:
            # Rejection sample until signs are compatible with an all 1 ground
            # state:
            signs = 2 * rng.integers(0, 1, endpoint=True, size=k) - 1
            
            
        # get the interactions for each clause
        for (u, usign), (v, vsign) in itertools.combinations(zip(variables, signs), 2):
            yield u, v, usign*vsign


def random_kmcsat(variables: typing.Union[int, typing.Sequence[dimod.typing.Variable]],
                  k: int,
                  num_clauses: int,
                  *,
                  plant_solution: bool = False,
                  seed: typing.Union[None, int, np.random.Generator] = None,
                  ) -> BinaryQuadraticModel:
    """Generate a random k Max-Cut satisfiability problem as a binary quadratic model.

    kMC-SAT [#ZK]_ is an NP-complete problem class
    that consists in satisfying a number of
    clauses of ``k`` literals (variables, or their negations).
    Each clause should encode a max-cut problem over the clause literals.
    
    Each clause contributes -:code:`k//2` to the energy when the clause is 
    satisfied, and at least 0 when unsatisfied. The energy :math:`H(s)` for a 
    spin assignment :math:`s` is thus lower bounded by :math:`E_{SAT}=-`:code:`k//2 * num_clauses`, 
    this lower bound matches the ground state energy in satisfiable instances. 
    For k>3, energy penalties per clause violation are non-uniform.


    Args:
        num_variables: The number of variables in the problem.
        k: number of variables participating in the clause.
        num_clauses: The number of clauses. Each clause contains three literals.
        plant_solution: Create literals uniformly subject to the constraint that the
            all 1 (and all -1) are ground states (satisfy all clauses).
        seed: Passed to :func:`numpy.random.default_rng()`, which is used
            to generate the clauses and the variable negations.
    Returns:
        A binary quadratic model with spin variables.

    .. note:: The clauses are randomly sampled from the space of k-variable
        clauses *with replacement* which can result in collisions. However,
        collisions are allowed in standard problem definitions, are absent with
        high probability in interesting cases, and are almost always harmless
        when they do occur. 

        For large problems planting of an all 1 solution can
        be achieved (in some special cases) without modification of the hardness
        qualities of the instance class. Planting of a not all 1 ground state
        can be achieved with a spin-reversal transform without loss of generality. [#DKR]_
    .. [#ZK] Lenka Zdeborová and Florent Krzakala,
       "Quiet Planting in the Locked Constraint Satisfaction Problems",
       https://epubs.siam.org/doi/10.1137/090750755
    .. [#DKR] Adam Douglass, Andrew D. King & Jack Raymond,
       "Constructing SAT Filters with a Quantum Annealer",
       https://link.springer.com/chapter/10.1007/978-3-319-24318-4_9
    """
    if isinstance(variables, collections.abc.Sequence):
        num_variables = len(variables)
        labels = variables
    else:
        num_variables = variables
        labels = None
        
    if num_variables < 1:
        raise ValueError("number of variables must be non-negative")
    elif k < 1:
        raise ValueError("number of variables must be non-negative")
    elif num_clauses < 0:
        raise ValueError("{num_clauses} must be non-negative")
    elif num_variables < k:
        raise ValueError(f"must use at least {k}<= number of variables")

    bqm = BinaryQuadraticModel(num_variables, Vartype.SPIN)
    bqm.add_quadratic_from(_kmcsat_interactions(num_variables, k, num_clauses,
                                                plant_solution=plant_solution, seed=seed))

    if labels:
        bqm.relabel_variables(dict(enumerate(labels)))

    return bqm


def random_nae3sat(variables: typing.Union[int, typing.Sequence[dimod.typing.Variable]],
                   num_clauses: int,
                   *,
                   plant_solution: bool = False,
                   seed: typing.Union[None, int, np.random.Generator] = None,
                   ) -> BinaryQuadraticModel:
    """Generate a random not-all-equal 3-satisfiability problem as a binary quadratic model.

    Not-all-equal 3-satisfiability (NAE3SAT_) is an NP-complete problem class
    that consists in satisfying a number of conjunctive
    clauses of three literals (variables, or their negations).
    For valid solutions, the literals in each clause should be not-all-equal;
    i.e. any assignment of values except ``(+1, +1, +1)`` or ``(-1, -1, -1)``
    are valid for each clause.

    Each clause contributes -1 to the energy when the clause is satisfied,
    and +3 when unsatisfied. The energy :math:`H(s)` for a spin assignment :math:`s`
    is thus lower bounded by :math:`E_{SAT}=-`:code:`num_clauses`, this lower 
    bound matches the ground state energy in satisfiable instances. The number 
    of violated clauses is :math:`(H(s) - E_{SAT})/4`.

    NAE3SAT problems have been studied with the D-Wave quantum annealer [#DKR]_.

    .. _NAE3SAT: https://en.wikipedia.org/wiki/Not-all-equal_3-satisfiability

    
    Args:
        num_variables: The number of variables in the problem.
        num_clauses: The number of clauses. Each clause contains three literals.
        plant_solution: Create literals uniformly subject to the constraint that the
            all 1 (and all -1) are ground states satisfying all clauses.
        seed: Passed to :func:`numpy.random.default_rng()`, which is used
            to generate the clauses and the variable negations.

    Returns:
        A binary quadratic model with spin variables.

    Example:

        Generate a NAE3SAT problem with a given clause-to-variable ratio (rho).

        >>> num_variables = 75
        >>> rho = 2.1
        >>> bqm = dimod.generators.random_nae3sat(num_variables, round(num_variables*rho))

    .. note:: The clauses are randomly sampled from the space of 3-variable
        clauses *with replacement* which can result in collisions. However,
        collisions are allowed in standard problem definitions, are absent with
        high probability in interesting cases, and are almost always harmless
        when they do occur. 

        Planting of a not all 1 solution state
        can be achieved with a spin-reversal transform without loss of 
        generality. Planting can significantly modify the hardness of 
        optimization problems.

    .. [#DKR] Adam Douglass, Andrew D. King & Jack Raymond,
       "Constructing SAT Filters with a Quantum Annealer",
       https://link.springer.com/chapter/10.1007/978-3-319-24318-4_9

    """
    return random_kmcsat(variables, 3, num_clauses, plant_solution=plant_solution, seed=seed)


def random_2in4sat(variables: typing.Union[int, typing.Sequence[dimod.typing.Variable]],
                   num_clauses: int,
                   *,
                   plant_solution: bool = False,
                   seed: typing.Union[None, int, np.random.Generator] = None,
                   ) -> BinaryQuadraticModel:
    """Generate a random 2-in-4 satisfiability problem as a binary quadratic model.

    2-in-4 satisfiability [#DKR]_ is an NP-complete problem class
    that consists in satisfying a number of conjunctive
    clauses of four literals (variables, or their negations).
    For valid solutions, two of the literals in each clause should ``+1`` and
    the other two should be ``-1``.

    Each clause contributes -2 to the energy when the clause is satisfied,
    and at least 0 when unsatisfied. The energy :math:`H(s)` for a spin 
    assignment :math:`s` is thus lower bounded by :math:`E_{SAT}=-2`:code:`num_clauses`, 
    this lower bound matches the ground state energy in satisfiable instances. 
    The number of violated clauses is at most :math:`(H(s) - E_{SAT})/2`.

    
    Args:
        num_variables: The number of variables in the problem.
        num_clauses: The number of clauses. Each clause contains three literals.
        plant_solution: Create literals uniformly subject to the constraint that the
            all 1 (and all -1) are ground states satisfying all clauses.
        seed: Passed to :func:`numpy.random.default_rng()`, which is used
            to generate the clauses and the variable negations.
    Returns:
        A binary quadratic model with spin variables.

    .. note:: The clauses are randomly sampled from the space of 4-variable
        clauses *with replacement* which can result in collisions. However,
        collisions are allowed in standard problem definitions, are absent with
        high probability in interesting cases, and are almost always harmless
        when they do occur. 

        Planting of a not all 1 ground state
        can be achieved with a spin-reversal transform without loss of 
        generality. Planting can significantly modify the hardness of 
        optimization problems. However, for large problems planting of a 
        solution can be achieved (in the SAT phase) without modification of the
        hardness qualities of the instance class. [#ZK]_ 
    .. [#DKR] Adam Douglass, Andrew D. King & Jack Raymond,
       "Constructing SAT Filters with a Quantum Annealer",
       https://link.springer.com/chapter/10.1007/978-3-319-24318-4_9
    .. [#ZK] Lenka Zdeborová and Florent Krzakala,
       "Quiet Planting in the Locked Constraint Satisfaction Problems",
       https://epubs.siam.org/doi/10.1137/090750755

    """
    return random_kmcsat(variables, 4, num_clauses, plant_solution=plant_solution, seed=seed)



# Copyright 2021 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import abc
import enum
import typing

from dimod.typing import Variable

__all__ = ['Sense', 'Eq', 'Ge', 'Le']


class Sense(enum.Enum):
    """Sense of a constraint.

    Supported values are:

    * ``Le``: less or equal (:math:`\le`)
    * ``Ge``: greater or equal (:math:`\ge`)
    * ``Eq``: equal (:math:`=`)

    Example:

        >>> from dimod import ConstrainedQuadraticModel, Integer
        >>> i = Integer("i")
        >>> cqm = ConstrainedQuadraticModel()
        >>> cqm.add_constraint(i <= 3, "i le 3")
        'i le 3'

    """

    Le = '<='
    Ge = '>='
    Eq = '=='


class Comparison(abc.ABC):
    def __init__(self, lhs, rhs):
        # todo: type checking
        self.lhs = lhs
        self.rhs = rhs

    def __init_subclass__(cls, sense=None, **kwargs):
        super().__init_subclass__(**kwargs)
        if sense is not None:
            cls.sense = sense if isinstance(sense, Sense) else Sense(sense)

    @property
    @abc.abstractmethod
    def sense(self):
        pass

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.lhs!r}, {self.rhs!r})"

    def __str__(self) -> str:
        return self.to_polystring()

    def to_polystring(self, encoder: typing.Optional[typing.Callable[[Variable], str]] = None) -> str:
        """Return a string representing the model as an equation.

        Args:
            encoder: A function mapping variables to a string. By default
                string variables are mapped directly whereas all other types
                are mapped to a string :code:`f"v{variable!r}"`.

        Returns:
            A string representing the comparison.

        """
        return f"{self.lhs.to_polystring(encoder)} {self.sense.value} {self.rhs!r}"


class Eq(Comparison, sense='=='):
    def __bool__(self):
        try:
            return self.lhs.is_equal(self.rhs)
        except AttributeError:
            return False


class Ge(Comparison, sense='>='):
    pass


class Le(Comparison, sense='<='):
    pass



# Copyright 2022 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from __future__ import annotations

import collections.abc
import functools
import io
import os
import shutil
import string
import tempfile
import typing

import dimod  # for typing

from dimod.cylp import cyread_lp_file
from dimod.sym import Sense
from dimod.vartypes import Vartype


__all__ = ['dump', 'dumps', 'loads', 'load']


LABEL_VALID_CHARS = set(string.ascii_letters + string.digits + "'!\"#$%&(),.;?@_‘’{}~")
LABEL_INVALID_FIRST_CHARS = set('eE.' + string.digits)


def _sign(bias: float) -> str:
    return '-' if bias < 0 else '+'


def _abs(bias: float) -> str:
    # todo: use bias.is_integer() instead of int(bias) == bias once we can require
    # numpy>=1.22.0.  this is needed to support Float32BQM on numpy<1.22.0.
    return repr(abs(int(bias))) if int(bias) == bias else repr(abs(bias))


def _sense(s: Sense) -> str:
    return '=' if s.value == "==" else s.value


def _label_error(label: typing.Hashable, rule: str) -> str:
    raise ValueError(f'Label {label!r} cannot be output to an LP file; {rule}. '
                     'Use CQM.relabel_variables() or CQM.relabel_constraints() to fix')


def _validate_label(label: typing.Hashable):
    if not isinstance(label, str):
        _label_error(label, 'labels must be strings')

    if not len(label):
        _label_error(label, 'labels must be nonempty')

    if len(label) > 255:
        _label_error(label, 'labels must be <= 255 characters in length')

    if not all(c in LABEL_VALID_CHARS for c in label):
        _label_error(label, f'labels must be composed of only these characters: '
                           f'{"".join(sorted(LABEL_VALID_CHARS))}')

    if any(label.startswith(c) for c in LABEL_INVALID_FIRST_CHARS):
        _label_error(label, f'labels must not start with any of these characters: '
                           f'{"".join(sorted(LABEL_INVALID_FIRST_CHARS))}')


class _WidthLimitedFile:
    TARGET_LINE_LEN = 80

    def __init__(self, fp: typing.TextIO):
        """Wrap a file-like object to provide a custom `write` method that
        inserts line breaks into the output stream.  This is done to generate attractive
        output (e.g. lines that are not extremely short) for most inputs, while satisfying
        the line length requirement of LP files for all inputs.

        This design has the advantage of a simple and readable implementation as compared
        to the approach of tracking line length and emitting newlines conditionally,
        through additional statements interspersed with the f.write() calls below.  And it
        has the advantage of producing more attractive output as compared to the dead
        simple approach of emitting newlines unconditionally in the f.write() calls below.

        This design does incur a run-time performance penalty of about 50% versus those
        alternatives.  In the balance of our goals, this seems to be the best option.
        """
        self.fp = fp

        self._line_len = 0

        self.tell = fp.tell

    def write(self, s: str):
        """Check if writing ``s`` would result in our line length target being
        exceeded, and if so, emit a newline.  Then write ``s``.

        Note that this function does not gaurantee a maximum line length.  Lines longer
        than the target length can be emitted.  Callers, i.e. the f.write() calls in the
        code below, must pass ``s`` arguments that are sufficiently short, to avoid
        exceeding the maximum line length for LP files.  The longest lines emitted by the
        current implementation are those encoding quadratic terms for an objective or
        constraints, when two variables have maximum length labels.  We have a unit test
        (test_long_line_breaking) to check this case.
        """
        try:
            pos = s.index('\n')
            rpos = s.rindex('\n')
        except ValueError:
            pos = len(s)
            rpos = None

        if self._line_len + pos > self.TARGET_LINE_LEN - 1:
            self.fp.write('\n ')
            self._line_len = 1

        self.fp.write(s)

        if rpos is not None:
            self._line_len = len(s) - 1 - rpos
        else:
            self._line_len += len(s)


def dump(cqm: dimod.ConstrainedQuadraticModel, file_like: typing.TextIO):
    """Serialize a constrained quadratic model as an LP file.

    LP files are a common format for encoding optimization models. See
    documentation from Gurobi_ and CPLEX_.

    Args:
        cqm: A constrained quadratic model.
        file_like: A ``.write()`` supporting file-like_ object.

    .. _file-like: https://docs.python.org/3/glossary.html#term-file-object

    .. _Gurobi: https://www.gurobi.com/documentation/9.5/refman/lp_format.html

    .. _CPLEX: https://www.ibm.com/docs/en/icos/12.8.0.0?topic=cplex-lp-file-format-algebraic-representation

    """
    # check that there are no soft constraints, LP format does not support them
    if len(cqm._soft) > 0:
        raise ValueError(f"LP file does not support soft constraints, {len(cqm._soft)} were given")

    # check that constraint labels are serializable
    for c in cqm.constraints:
        _validate_label(c)

    vartypes = {var: cqm.vartype(var) for var in cqm.variables}

    # check that variable labels and types are serializable
    for v, vartype in vartypes.items():
        _validate_label(v)

        if vartype == Vartype.SPIN:
            raise ValueError(
                'SPIN variables not supported in LP files, convert them to BINARY beforehand.')

    f = _WidthLimitedFile(file_like)

    # write the objective
    for var, bias in cqm.objective.iter_linear():
        if bias:
            if not f.tell():
                # LP files allow to omit this part if the objective is empty, so
                # we add these lines if we are sure that there are nonzero terms.
                f.write("Minimize\n")
                f.write(' obj: ')
            f.write(f"{_sign(bias)} {_abs(bias)} {var} ")

    if not cqm.objective.is_linear():
        if not f.tell():
            # if the objective is only quadratic then the header is written now
            f.write("Minimize\n")
            f.write(' obj: ')

        f.write('+ [ ')
        for u, v, bias in cqm.objective.iter_quadratic():
            # multiply bias by two because all quadratic terms are eventually
            # divided by two outside the squared parenthesis
            f.write(f"{_sign(bias)} {_abs(2 * bias)} {u} * {v} ")
        f.write(']/2 ')

    if cqm.objective.offset:
        offset = cqm.objective.offset

        if not f.tell():
            # if the objective has only an offset then the header is written now
            f.write("Minimize\n")
            f.write(' obj: ')
        f.write(f"{_sign(offset)} {_abs(offset)} ")

    # write the constraints
    f.write("\n\n")
    f.write("Subject To \n")

    for label, constraint in cqm.constraints.items():
        f.write(f' {label}: ')

        for var, bias in constraint.lhs.iter_linear():
            if bias:
                f.write(f"{_sign(bias)} {_abs(bias)} {var} ")

        if constraint.lhs.quadratic:
            f.write('+ [ ')
            for u, v, bias in constraint.lhs.iter_quadratic():
                f.write(f"{_sign(bias)} {_abs(bias)} {u} * {v} ")
            f.write('] ')

        rhs = constraint.rhs - constraint.lhs.offset
        f.write(f" {_sense(constraint.sense)} {rhs}\n")

    # write variable bounds
    f.write('\n')
    f.write('Bounds\n')

    for v, vartype in vartypes.items():
        if vartype in (Vartype.INTEGER, Vartype.REAL):
            f.write(f' {cqm.lower_bound(v)} <= {v} <= {cqm.upper_bound(v)}\n')
        elif vartype is not Vartype.BINARY:
            raise RuntimeError(f"unexpected vartype {vartype}")

    # write variable names
    for section, vartype_ in (('Binary', Vartype.BINARY), ('General', Vartype.INTEGER)):
        f.write('\n')
        f.write(f'{section}\n')

        for v, vartype in vartypes.items():
            if vartype is vartype_:
                f.write(f' {v}')

    # conclude
    f.write('\n')
    f.write('End')
    return f


def dumps(cqm: dimod.ConstrainedQuadraticModel) -> str:
    """Serialize a constrained quadratic model as an LP file.

    LP files are a common format for encoding optimization models. See
    documentation from Gurobi_ and CPLEX_.

    Args:
        cqm: A constrained quadratic model.

    Returns:
        A string encoding the constrained quadratic model as an LP file.

    .. _Gurobi: https://www.gurobi.com/documentation/9.5/refman/lp_format.html

    .. _CPLEX: https://www.ibm.com/docs/en/icos/12.8.0.0?topic=cplex-lp-file-format-algebraic-representation

    """
    with io.StringIO() as f:
        dump(cqm, f)
        f.seek(0)
        return f.read()


def load(file_like: typing.Union[str, bytes, io.IOBase]) -> dimod.ConstrainedQuadraticModel:
    """Construct a constrained quadratic model from a LP file.

    LP files are a common format for encoding optimization models. See
    documentation from Gurobi_ and CPLEX_.

    Note that if the objective function is specified as a maximization function
    then it will be converted to a minimization function by flipping the sign
    of all of the biases.

    Args:
        file_like: Either a :class:`str` or :class:`bytes` object representing
            a path, or a file-like_ object.

    Returns:

        An example of reading from a LP file.

        .. code-block:: python

            with open('example.lp', 'rb') as f:
                cqm = dimod.lp.load(f)

        An example of loading from a path

        .. code-block:: python

            cqm = dimod.lp.load('example.lp')

    See also:

        :func:`~dimod.serialization.lp.loads`

    .. versionadded:: 0.11.0

    .. _Gurobi: https://www.gurobi.com/documentation/9.5/refman/lp_format.html

    .. _CPLEX: https://www.ibm.com/docs/en/icos/12.8.0.0?topic=cplex-lp-file-format-algebraic-representation

    .. _file-like: https://docs.python.org/3/glossary.html#term-file-object

    """

    if isinstance(file_like, (str, bytes)):
        return cyread_lp_file(file_like)

    # ok, we got a file-like

    if not file_like.readable():
        raise ValueError("file_like must be readable")

    if file_like.seekable() and file_like.tell():
        # this file is current pointing somewhere other than the beginning,
        # so we make a copy so our reader can start at the beginning
        filename = ''
    else:
        try:
            filename = file_like.name
        except AttributeError:
            filename = ''

    if (filename is not None) and os.path.isfile(filename):
        return cyread_lp_file(filename)        

    # copy it into somewhere that our reader can get it
    with tempfile.NamedTemporaryFile('wb', delete=False) as tf:
        shutil.copyfileobj(file_like, tf)

    try:
        return cyread_lp_file(tf.name)
    finally:
        # remove the file so we're not accumulating memory/disk space
        os.unlink(tf.name)


def loads(obj: typing.Union[str, typing.ByteString]) -> dimod.ConstrainedQuadraticModel:
    """Construct a constrained quadratic model from a string formatted as a LP file.

    LP files are a common format for encoding optimization models. See
    documentation from Gurobi_ and CPLEX_.

    Note that if the objective function is specified as a maximization function
    then it will be converted to a minimization function by flipping the sign
    of all of the biases.

    Args:
        obj: A :class:`str` or :class:`bytes` formatted like an LP file.

    Returns:

        An example of reading a string formatted as an LP file.

        .. testcode::

            lp = '''
            Minimize
                x0 - 2 x1
            Subject To
                x0 + x1 = 1
            Binary
                x0 x1
            End
            '''

            cqm = dimod.lp.loads(lp)


    See also:

        :func:`~dimod.serialization.lp.load`

    .. versionadded:: 0.11.0

    .. _Gurobi: https://www.gurobi.com/documentation/9.5/refman/lp_format.html

    .. _CPLEX: https://www.ibm.com/docs/en/icos/12.8.0.0?topic=cplex-lp-file-format-algebraic-representation

    .. _file-like: https://docs.python.org/3/glossary.html#term-file-object

    """

    if isinstance(obj, str):
        obj = obj.encode()

    return load(io.BytesIO(obj))



# Copyright 2022 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from __future__ import annotations

import abc
import numbers
import tempfile
import typing

import numpy as np

import dimod

from dimod.constrained.cyexpression import cyObjectiveView, cyConstraintView
from dimod.views.quadratic import QuadraticViewsMixin


class _ExpressionMixin(QuadraticViewsMixin):
    # dev note: I think most of these can actually be promoted to the QuadraticViewsMixin
    # one, but let's leave it alone for now

    @property
    @abc.abstractmethod
    def dtype(self) -> np.dtype:
        pass

    @property
    @abc.abstractmethod
    def num_variables(self) -> int:
        pass

    @property
    def shape(self) -> typing.Tuple[int, int]:
        return self.num_variables, self.num_interactions

    def energy(self, sample, dtype=None) -> dimod.typing.Bias:
        energies = self.energies(sample)

        if not len(energies):  # the empty case, happens with []
            return self.dtype.type(0)

        energy, = energies
        return energy

    @abc.abstractmethod
    def energies(self, samples_like):
        pass

    # developer note: in the future we should just generalize these to QuadraticViewsMixin
    # todo: __mul__ and __div__, though they don't see much use so will leave that alone for now
    def __add__(self, other):
        if isinstance(other, (QuadraticViewsMixin, numbers.Number)):
            qm = dimod.QuadraticModel()
            qm.update(self)
            qm += other
            return qm

        return NotImplemented

    def __radd__(self, other):
        if isinstance(other, (QuadraticViewsMixin, numbers.Number)):
            qm = dimod.QuadraticModel()
            qm.update(self)
            return other + qm

        return NotImplemented

    def __sub__(self, other):
        if isinstance(other, (QuadraticViewsMixin, numbers.Number)):
            qm = dimod.QuadraticModel()
            qm.update(self)
            qm -= other
            return qm

        return NotImplemented

    def __rsub__(self, other):
        if isinstance(other, (QuadraticViewsMixin, numbers.Number)):
            qm = dimod.QuadraticModel()
            qm.update(self)
            return other - qm

        return NotImplemented

    def is_almost_equal(self, other, places: int = 7) -> bool:
        """Test for near equality to all biases of a given quadratic model.

        Args:
            other:
                Quadratic model with which to compare biases.
            places:
                Number of decimal places to which the Python :func:`round`
                function calculates approximate equality.

        Examples:
            >>> from dimod import QuadraticModel
            >>> qm1 = QuadraticModel({'x': 0.0, 'i': 0.1234}, {('i', 'x'): -1.1234},
            ...                      0.0, {'x': 'BINARY', 'i': 'INTEGER'})
            >>> qm2 = QuadraticModel({'x': 0.0, 'i': 0.1232}, {('i', 'x'): -1.1229},
            ...                      0.0, {'x': 'BINARY', 'i': 'INTEGER'})
            >>> qm1.is_almost_equal(qm2, 4)
            False
            >>> qm1.is_almost_equal(qm2, 3)
            True
        """
        if isinstance(other, numbers.Number):
            return not (self.num_variables or round(self.offset - other, places))

        def eq(a, b):
            return not round(a - b, places)

        try:
            if callable(other.vartype):
                vartype_eq = all(self.vartype(v) == other.vartype(v) for v in self.variables)
            else:
                vartype_eq = all(self.vartype(v) == other.vartype for v in self.variables)

            return (vartype_eq
                    and self.shape == other.shape
                    and eq(self.offset, other.offset)
                    and all(eq(self.get_linear(v), other.get_linear(v))
                            for v in self.variables)
                    and all(eq(bias, other.get_quadratic(u, v))
                            for u, v, bias in self.iter_quadratic())
                    )
        except (AttributeError, ValueError):
            # it's not a BQM or variables/interactions don't match
            return False

    def is_equal(self, other) -> bool:
        """Return True if the given model has the same variables, vartypes and biases.

        Args:
            other: Quadratic model to compare against.
        """
        if isinstance(other, numbers.Number):
            return not self.num_variables and bool(self.offset == other)
        # todo: performance

        try:
            if callable(other.vartype):
                vartype_eq = all(self.vartype(v) == other.vartype(v) for v in self.variables)
            else:
                vartype_eq = all(self.vartype(v) == other.vartype for v in self.variables)

            return (vartype_eq
                    and self.shape == other.shape  # redundant, fast to check
                    and self.offset == other.offset
                    and self.linear == other.linear
                    and self.adj == other.adj)
        except AttributeError:
            return False

    def to_file(self, *, spool_size: int = int(1e9)) -> tempfile.SpooledTemporaryFile:
        # serialize as a QM.
        # todo: turn this into a function rather than a method
        from dimod.quadratic.quadratic_model import QuadraticModel
        from dimod.variables import Variables

        class Model:
            data = self
            dtype = self.dtype
            index_dtype = self.index_dtype
            num_variables = self.num_variables
            shape = (self.num_variables, self.num_interactions)
            variables = Variables(self.variables)

        return QuadraticModel.to_file(Model(), spool_size=spool_size)

    @abc.abstractmethod
    def vartype(self, v: dimod.typing.Variable) -> dimod.Vartype:
        pass


class ObjectiveView(cyObjectiveView, _ExpressionMixin):
    pass


class ConstraintView(cyConstraintView, _ExpressionMixin):
    pass
